package mypkg.lambda.util;

import java.io.Reader;
import java.io.FilterReader;
import java.io.BufferedReader;
import java.io.IOException;

/**
 * Rg폜 Reader ̃TuNX
 */
public class CommentRemover extends FilterReader {

	private BufferedReader in;
	private char buffer[];
	private int rptr;
	private int endptr;
	private boolean verbose;

	private final String blockComment1;
	private final String blockComment2;
	private final String lineComment;

	/**
	 * RXgN^
	 *
	 * @param in ͗p Reader
	 */
	public CommentRemover(Reader in, String blockComment1, String blockComment2,
			String lineComment, boolean verbose) {
		super(in);
		if (blockComment1.length() != 2 || blockComment2.length() != 2
				|| lineComment.length() != 2)
			throw new IllegalArgumentException("Sorry, length must be 2.");
		this.blockComment1 = blockComment1;
		this.blockComment2 = blockComment2;
		this.lineComment = lineComment;
		this.in = (in instanceof BufferedReader)
			? (BufferedReader)in : new BufferedReader(in);
		buffer = new char[2048];
		rptr = endptr = 0;
		this.verbose = verbose;
	}

	/**
	 * N[Y
	 */
	public void close() throws IOException {
		if (in != null) {
			in.close();
			in = null;
			buffer = null;
			if (verbose)
				System.out.println("# "+getClass().getName()+".close()");
		}
	}

	/**
	 * ǂݍ񂾂P߂lŕԂ.
	 */
	public int read() throws IOException {
		if (rptr >= endptr)
			if (fillbuf() <= 0)
				return -1;
		return buffer[rptr++] & 0xffff;
	}

	/**
	 * w̃obt@Ƀf[^ǂݍ.
	 */
	public int read(char cbuf[], int off, int len) throws IOException {
		int start = off;
		for (int end = off + len; off < end; ) {
			int c = read();
			if (c < 0)
				break;
			cbuf[off++] = (char)c;
			if (c == '\n' && !in.ready())  //TODO: Mac
				break;
		}
		return (off > start)? off - start : -1;
	}

	private int fillbuf() throws IOException {
		rptr = endptr = 0;
		for (int end = buffer.length - 8; endptr < end; ) {
			int c = in.read();
			if (c < 0)
				break;
			if (c == '\n') {  //TODO: Mac
				buffer[endptr++] = (char)c;
				break;
			}
			if (c == '\'' || c == '"') {
				// '"', "//XXX", "/*YYY*/" ̉\r
				if (handleQuote(c))
					continue;
				break;
			}
			if (c != lineComment.charAt(0) && c != blockComment1.charAt(0)) {
				buffer[endptr++] = (char)c;
				continue;
			}
			int c2 = in.read();
			if (c2 < 0) {
				buffer[endptr++] = (char)c;
				break;
			}
			if (isLineComment(c, c2)) {
				singleLineComment();
			} else if (isBlockCommentStart(c, c2)) {
				multiLineComment();
			} else {
				buffer[endptr++] = (char)c;
				buffer[endptr++] = (char)c2;
			}
		}
		return endptr;
	}

	/**
	 * sRg𔼊pXy[Xɒu.
	 * 폜ł͂Ȃu闝R̓G[bZ[W̍sƃJۑ
	 * ߂ł.
	 */
	private void multiLineComment() throws IOException {
		if (endptr >= buffer.length-1)
			expand();
		buffer[endptr++] = (char)' ';
		buffer[endptr++] = (char)' ';
		for (int c1 = -1, c2 = -1; ; c1 = c2) {
			if ((c2 = in.read()) < 0)
				throw new IOException("invalid comment");
			if (endptr >= buffer.length-1)
				expand();
			if (c2 == '\n' || c2 == '\r') {
				buffer[endptr++] = (char)c2;
			} else {
				buffer[endptr++] = (char)' ';
				if (fullWidth(c2))
					buffer[endptr++] = (char)' ';
			}
			if (isBlockCommentEnd(c1, c2))
				break;
		}
	}

	private void singleLineComment() throws IOException {
		int c = -1;
		while ((c = in.read()) >= 0 && c != '\r' && c != '\n')
			;
		if (c >= 0)
			buffer[endptr++] = (char)c;
	}

	private boolean handleQuote(int qc) throws IOException {
		// obt@sƂȂgđΉ.
		buffer[endptr++] = (char)qc;
		boolean escape = false;
		for ( ; ; ) {
			int c = in.read();
			if (c < 0)
				throw new IOException("invalid quotation");
			if (endptr >= buffer.length)
				expand();
			buffer[endptr++] = (char)c;
			if (escape)
				escape = false;
			else if (c == '\\')
				escape = true;
			else if (c == qc)
				break;
			else if (c == '\n')  //TODO: Mac
				return false;    // ERROR
		}
		return true;
	}

	private void expand() {
		char newbuf[] = new char[buffer.length + 256];
		System.arraycopy(buffer, 0, newbuf, 0, endptr);
		buffer = newbuf;
	}

	private boolean fullWidth(int c) {
		return c >= '\u0080' && !('\uff61' <= c && c <= '\uff9f');
	}

	private boolean isLineComment(int c1, int c2) {
		return c1 == lineComment.charAt(0) && c2 == lineComment.charAt(1);
	}

	private boolean isBlockCommentStart(int c1, int c2) {
		return c1 == blockComment1.charAt(0) && c2 == blockComment1.charAt(1);
	}

	private boolean isBlockCommentEnd(int c1, int c2) {
		return c1 == blockComment2.charAt(0) && c2 == blockComment2.charAt(1);
	}
}
